# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Custom stylesheets used in Spyder."""

# Standard library imports
import copy
import os
import sys

# Third-party imports
import qdarkstyle
from qstylizer.parser import parse as parse_stylesheet
import qstylizer.style

# Local imports
from spyder.api.config.mixins import SpyderConfigurationAccessor
from spyder.api.fonts import SpyderFontType, SpyderFontsMixin
from spyder.api.utils import classproperty
from spyder.config.gui import is_dark_interface
from spyder.utils.palette import SpyderPalette


# =============================================================================
# ---- Constants
# =============================================================================
MAC = sys.platform == 'darwin'
WIN = os.name == 'nt'


class AppStyle(SpyderFontsMixin):
    """Enum with several constants used in the application style."""

    # Size of margins.
    MarginSize = 3  # px

    # Size of find widget line edits (e.g. FinderWidget and FindReplace)
    FindMinWidth = 400  # px
    FindHeight = 26  # px

    # To have it for quick access because it's needed a lot in Mac
    MacScrollBarWidth = 16  # px

    # Icon size in config pages
    ConfigPageIconSize = 20

    # Padding for QPushButton's
    QPushButtonPadding = f'{MarginSize + 1}px {4 * MarginSize}px'

    @classproperty
    def _fs(cls):
        """Interface font size in points."""
        return cls.get_font(SpyderFontType.Interface).pointSize()

    @classproperty
    def ComboBoxMinHeight(cls):
        """Combobox min height in em's."""
        font_size = cls._fs

        if font_size < 10:
            min_height = 1.8
        elif 10 <= font_size < 13:
            min_height = 1.7 if MAC else 1.6
        else:
            min_height = 1.5 if MAC else 1.4

        return min_height

    # Padding for content inside an element of higher hierarchy
    InnerContentPadding = 5 * MarginSize


# =============================================================================
# ---- Base stylesheet class
# =============================================================================
class SpyderStyleSheet:
    """Base class for Spyder stylesheets."""

    SET_STYLESHEET_AT_INIT = True
    """
    Decide if the stylesheet must be set when the class is initialized.

    Notes
    -----
    There are some stylesheets for which this is not possible (e.g. the ones
    that need to access our fonts).
    """

    def __init__(self):
        self._stylesheet = qstylizer.style.StyleSheet()
        if self.SET_STYLESHEET_AT_INIT:
            self.set_stylesheet()

    def get_stylesheet(self):
        return self._stylesheet

    def to_string(self):
        if self._stylesheet.toString() == "":
            self.set_stylesheet()
        return self._stylesheet.toString()

    def get_copy(self):
        """
        Return a copy of the sytlesheet.

        This allows it to be modified for specific widgets.
        """
        if self._stylesheet.toString() == "":
            self.set_stylesheet()
        return copy.deepcopy(self)

    def set_stylesheet(self):
        raise NotImplementedError(
            "Subclasses need to implement this method to set the _stylesheet "
            "attribute as a Qstylizer StyleSheet object."
        )

    def __str__(self):
        """
        Get a string representation of the stylesheet object this class
        holds.
        """
        return self.to_string()


# =============================================================================
# ---- Application stylesheet
# =============================================================================
class AppStylesheet(SpyderStyleSheet, SpyderConfigurationAccessor):
    """
    Class to build and access the stylesheet we use in the entire
    application.
    """

    # Don't create the stylesheet here so that Spyder gets the app font from
    # the system when it starts for the first time. This also allows us to
    # display the splash screen more quickly because the stylesheet is then
    # computed only when it's going to be applied to the app, not when this
    # object is imported.
    SET_STYLESHEET_AT_INIT = False

    def __init__(self):
        super().__init__()
        self._stylesheet_as_string = None

    def to_string(self):
        """Save stylesheet as a string for quick access."""
        if self._stylesheet_as_string is None:
            self.set_stylesheet()
            self._stylesheet_as_string = self._stylesheet.toString()
        return self._stylesheet_as_string

    def set_stylesheet(self):
        """
        This takes the stylesheet from QDarkstyle and applies our
        customizations to it.
        """
        stylesheet = qdarkstyle.load_stylesheet(palette=SpyderPalette)
        self._stylesheet = parse_stylesheet(stylesheet)

        # Add our customizations
        self._customize_stylesheet()

    def _customize_stylesheet(self):
        """Apply our customizations to the stylesheet."""
        css = self._stylesheet

        # App font properties
        font_family = self.get_conf('app_font/family', section='appearance')
        font_size = int(self.get_conf('app_font/size', section='appearance'))

        # Remove padding and border for QStackedWidget (used in Plots
        # and the Variable Explorer)
        css['QStackedWidget'].setValues(
            border='0px',
            padding='0px',
        )

        # Remove margin when pressing buttons
        css["QToolButton:pressed"].setValues(
            margin='0px'
        )

        # Remove border, padding and spacing for main toolbar
        css.QToolBar.setValues(
            borderBottom='0px',
            padding='0px',
            spacing='0px',
        )

        # Remove margins around separators and decrease size a bit
        css['QMainWindow::separator:horizontal'].setValues(
            marginTop='0px',
            marginBottom='0px',
            # This is summed to the separator padding (2px)
            width="3px",
            # Hide image because the default image is not visible at this size
            image="none"
        )

        css['QMainWindow::separator:vertical'].setValues(
            marginLeft='0px',
            marginRight='0px',
            # This is summed to the separator padding (2px)
            height='3px',
            # Hide image because the default image is not visible at this size
            image="none"
        )

        # Increase padding and fix disabled color for QPushButton's
        css.QPushButton.setValues(padding=AppStyle.QPushButtonPadding)

        for state in ['disabled', 'checked', 'checked:disabled']:
            css[f'QPushButton:{state}'].setValues(
                padding=AppStyle.QPushButtonPadding,
            )

            # This is especially necessary in the light theme because the
            # contrast between the background and text colors is too small
            if state in ['disabled', 'checked:disabled']:
                css[f"QPushButton:{state}"].setValues(
                    color=SpyderPalette.COLOR_TEXT_3,
                )

        # Adjust QToolButton style to our needs.
        # This affects not only the pane toolbars but also the
        # find/replace widget, the finder in the Variable Explorer,
        # and all QToolButton's that are not part of the main toolbar.
        for element in ['QToolButton', 'QToolButton:disabled']:
            css[f'{element}'].setValues(
                backgroundColor='transparent'
            )

        for state in ['hover', 'pressed', 'checked', 'checked:hover']:
            if state == 'hover':
                color = SpyderPalette.COLOR_BACKGROUND_2
            else:
                color = SpyderPalette.COLOR_BACKGROUND_3
            css[f'QToolButton:{state}'].setValues(
                backgroundColor=color
            )

        # Adjust padding of QPushButton's in QDialog's
        for widget in ["QPushButton", "QPushButton:disabled"]:
            css[f"QDialogButtonBox {widget}"].setValues(
                padding=(
                    AppStyle.QPushButtonPadding
                    if (MAC or WIN)
                    else
                    f"{AppStyle.MarginSize + 1}px {AppStyle.MarginSize}px"
                ),
                # This width comes from QDarkstyle but it's too big on Mac
                minWidth="50px" if WIN else ("60px" if MAC else "80px"),
            )

        css["QDialogButtonBox QPushButton:!default"].setValues(
            padding=(
                AppStyle.QPushButtonPadding
                if (MAC or WIN)
                else
                f"{AppStyle.MarginSize + 1}px {AppStyle.MarginSize}px"
            ),
            # This width comes from QDarkstyle but it's too big on Mac
            minWidth="50px" if WIN else ("60px" if MAC else "80px"),
        )

        # Remove icons in QMessageBoxes
        css["QDialogButtonBox"]["dialogbuttonbox-buttons-have-icons"].setValue(
            "0"
        )

        # Set font for widgets that don't inherit it from the application
        # This is necessary for spyder-ide/spyder#5942.
        for widget in ['QToolTip', 'QDialog', 'QListView', 'QTreeView',
                       'QHeaderView::section', 'QTableView']:
            css[f'{widget}'].setValues(
                fontFamily=font_family,
                fontSize=f'{font_size}pt'
            )

        # Make lineedits have *almost* the same height as our comboboxes. This
        # is not perfect because (oddly enough) Qt doesn't set the same height
        # for both when using the same value, but it's close enough.
        css.QLineEdit.setValues(
            minHeight=f'{AppStyle.ComboBoxMinHeight - 0.25}em'
        )

        # Do the same for spinboxes
        css.QSpinBox.setValues(
            minHeight=f'{AppStyle.ComboBoxMinHeight - 0.25}em'
        )

        # Remove border in QGroupBox to avoid the "boxes within boxes"
        # antipattern. Also, increase its title font in one point to make it
        # more relevant.
        css.QGroupBox.setValues(
            border='0px',
            fontSize=f'{font_size + 1}pt',
        )

        # Increase separation between title and content of QGroupBoxes and fix
        # its alignment.
        css['QGroupBox::title'].setValues(
            paddingTop='-0.3em',
            left='0px',
        )

        # Decrease splitter handle size to be a bit smaller than QMainWindow
        # separators.
        css['QSplitter::handle'].setValues(
            padding="0px",
        )

        css['QSplitter::handle:horizontal'].setValues(
            width="5px",
            image="none"
        )

        css['QSplitter::handle:vertical'].setValues(
            height="5px",
            image="none"
        )

        # Make splitter handle color match the one of QMainWindow separators
        css['QSplitter::handle:hover'].setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_6,
        )

        # Add padding to tooltips
        css.QToolTip.setValues(
            padding="1px 2px",
        )

        # Add padding to tree widget items to make them look better
        css["QTreeWidget::item"].setValues(
            padding=f"{AppStyle.MarginSize - 1}px 0px",
        )

        css["QTreeView::item"].setValues(
            padding=f"{AppStyle.MarginSize - 1}px 0px",
        )


APP_STYLESHEET = AppStylesheet()

# =============================================================================
# ---- Toolbar stylesheets
# =============================================================================
class ApplicationToolbarStylesheet(SpyderStyleSheet):
    """Stylesheet for application toolbars."""

    BUTTON_WIDTH = '47px'
    BUTTON_HEIGHT = '47px'
    BUTTON_MARGIN_LEFT = '3px'
    BUTTON_MARGIN_RIGHT = '3px'

    def set_stylesheet(self):
        css = self.get_stylesheet()

        # Main background color
        css.QToolBar.setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_4
        )

        # Adjust QToolButton to follow the main toolbar style.
        css.QToolButton.setValues(
            width=self.BUTTON_WIDTH,
            height=self.BUTTON_HEIGHT,
            marginLeft=self.BUTTON_MARGIN_RIGHT,
            marginRight=self.BUTTON_MARGIN_RIGHT,
            border='0px',
            borderRadius='0px',
            padding='0px',
        )

        for state in ['hover', 'pressed', 'checked', 'checked:hover']:
            if state == 'hover':
                color = SpyderPalette.COLOR_BACKGROUND_5
            else:
                color = SpyderPalette.COLOR_BACKGROUND_6
            css[f'QToolBar QToolButton:{state}'].setValues(
                backgroundColor=color
            )

        # Remove indicator for popup mode
        css['QToolBar QToolButton::menu-indicator'].setValues(
            image='none'
        )


class PanesToolbarStyleSheet(SpyderStyleSheet):
    """Stylesheet for pane toolbars."""

    # These values make buttons to be displayed at 44px according to Gammaray
    BUTTON_WIDTH = '37px'
    BUTTON_HEIGHT = '37px'

    def set_stylesheet(self):
        css = self.get_stylesheet()

        css.QToolBar.setValues(
            spacing='4px'
        )

        css.QToolButton.setValues(
            height=self.BUTTON_HEIGHT,
            width=self.BUTTON_WIDTH,
            border='0px',
            borderRadius='0px',
            margin='0px'
        )

        # Remove indicator for popup mode
        css['QToolButton::menu-indicator'].setValues(
            image='none'
        )


APP_TOOLBAR_STYLESHEET = ApplicationToolbarStylesheet()
PANES_TOOLBAR_STYLESHEET = PanesToolbarStyleSheet()


# =============================================================================
# ---- Tabbar stylesheets
# =============================================================================
class BaseTabBarStyleSheet(SpyderStyleSheet):
    """Base style for tabbars."""

    OBJECT_NAME = ''

    # Additional border for scroll buttons
    SCROLL_BUTTONS_BORDER_WIDTH = '0px'

    # Position for the scroll buttons additional border
    SCROLL_BUTTONS_BORDER_POS = ''

    def set_stylesheet(self):
        css = self.get_stylesheet()
        buttons_color = SpyderPalette.COLOR_BACKGROUND_1

        # Set style for scroll buttons
        css[f'QTabBar{self.OBJECT_NAME} QToolButton'].setValues(
            background=buttons_color,
            borderRadius='0px',
        )

        if self.SCROLL_BUTTONS_BORDER_POS == 'right':
            css[f'QTabBar{self.OBJECT_NAME} QToolButton'].setValues(
                borderRight=(
                    f'{self.SCROLL_BUTTONS_BORDER_WIDTH} solid {buttons_color}'
                )
            )
        else:
            css[f'QTabBar{self.OBJECT_NAME} QToolButton'].setValues(
                borderBottom=(
                    f'{self.SCROLL_BUTTONS_BORDER_WIDTH} solid {buttons_color}'
                )
            )

        # Hover and pressed state for scroll buttons
        for state in ['hover', 'pressed', 'checked', 'checked:hover']:
            if state == 'hover':
                color = SpyderPalette.COLOR_BACKGROUND_2
            else:
                color = SpyderPalette.COLOR_BACKGROUND_3
            css[f'QTabBar{self.OBJECT_NAME} QToolButton:{state}'].setValues(
                background=color
            )

        # Set width for scroll buttons
        css['QTabBar::scroller'].setValues(
            width='66px',
        )


class PanesTabBarStyleSheet(PanesToolbarStyleSheet, BaseTabBarStyleSheet):
    """Stylesheet for pane tabbars"""

    TOP_MARGIN = '12px'
    OBJECT_NAME = '#pane-tabbar'
    SCROLL_BUTTONS_BORDER_WIDTH = '5px'
    SCROLL_BUTTONS_BORDER_POS = 'right'

    def set_stylesheet(self):
        # Calling super().set_stylesheet() here doesn't work.
        PanesToolbarStyleSheet.set_stylesheet(self)
        BaseTabBarStyleSheet.set_stylesheet(self)
        css = self.get_stylesheet()

        # This removes a white dot that appears to the left of right corner
        # widgets
        css.QToolBar.setValues(
            marginLeft='-3px' if WIN else '-1px',
        )

        # QTabBar forces the corner widgets to be smaller than they should.
        # be. The added top margin allows the toolbuttons to expand to their
        # normal size.
        # See: spyder-ide/spyder#13600
        css['QTabBar::tab'].setValues(
            marginTop=self.TOP_MARGIN,
            paddingTop='4px',
            paddingBottom='4px',
            paddingLeft='4px' if MAC else '10px',
            paddingRight='10px' if MAC else '4px'
        )

        if MAC:
            # Show tabs left-aligned on Mac and remove spurious
            # pixel to the left.
            css.QTabBar.setValues(
                alignment='left',
                marginLeft='-1px'
            )

            css['QTabWidget::tab-bar'].setValues(
                alignment='left',
            )
        else:
            # Remove spurious pixel to the left
            css.QTabBar.setValues(
                marginLeft='-3px' if WIN else '-1px'
            )

        # Fix minor visual glitch when hovering tabs
        # See spyder-ide/spyder#15398
        css['QTabBar::tab:hover'].setValues(
            paddingTop='3px',
            paddingBottom='3px',
            paddingLeft='3px' if MAC else '9px',
            paddingRight='9px' if MAC else '3px'
        )

        for state in ['selected', 'selected:hover']:
            css[f'QTabBar::tab:{state}'].setValues(
                paddingTop='4px',
                paddingBottom='3px',
                paddingLeft='4px' if MAC else '10px',
                paddingRight='10px' if MAC else '4px'
            )

        # Remove border between selected tab and pane below
        css['QTabWidget::pane'].setValues(
            borderTop='0px',
        )

        # Adjust margins of corner widgets
        css['QTabWidget::left-corner'].setValues(
            top='-1px',
        )

        css['QTabWidget::right-corner'].setValues(
            top='-1px',
            right='-3px' if WIN else '-1px'
        )

        # Make scroll buttons height match the one of tabs
        css[f'QTabBar{self.OBJECT_NAME} QToolButton'].setValues(
            marginTop=self.TOP_MARGIN,
        )

        # Make scroll button icons smaller on Windows and Mac
        if WIN or MAC:
            css[f'QTabBar{self.OBJECT_NAME} QToolButton'].setValues(
                padding=f'{5 if WIN else 7}px',
            )


class BaseDockTabBarStyleSheet(BaseTabBarStyleSheet):
    """Base style for dockwidget tabbars."""

    SCROLL_BUTTONS_BORDER_WIDTH = '2px'
    SCROLL_BUTTONS_PADDING = 7 if WIN else 9

    def set_stylesheet(self):
        super().set_stylesheet()

        # Main constants
        css = self.get_stylesheet()

        # Center tabs to differentiate them from the regular ones.
        # See spyder-ide/spyder#9763 for details.
        css.QTabBar.setValues(
            alignment='center'
        )

        css['QTabWidget::tab-bar'].setValues(
            alignment='center'
        )

        # Style for selected tabs
        css['QTabBar::tab:selected'].setValues(
            color=(
                SpyderPalette.COLOR_TEXT_1 if is_dark_interface() else
                SpyderPalette.COLOR_BACKGROUND_1
            ),
            backgroundColor=SpyderPalette.SPECIAL_TABS_SELECTED,
        )

        # Make scroll button icons smaller on Windows and Mac
        if WIN or MAC:
            css['QTabBar QToolButton'].setValues(
                padding=f'{self.SCROLL_BUTTONS_PADDING}px',
            )


class SpecialTabBarStyleSheet(BaseDockTabBarStyleSheet):
    """
    Style for special tab bars.

    Notes
    -----
    This is the base class for horizontal tab bars that follow the design
    discussed on issue spyder-ide/ux-improvements#4.
    """

    SCROLL_BUTTONS_BORDER_POS = 'right'

    def set_stylesheet(self):
        super().set_stylesheet()

        # -- Main constants
        css = self.get_stylesheet()
        margin_size = AppStyle.MarginSize

        # -- Basic style
        css['QTabBar::tab'].setValues(
            # Only add margin to the bottom
            margin=f'0px 0px {2 * margin_size}px 0px',
            # Border radius is added for specific tabs (see below)
            borderRadius='0px',
            # Remove a colored border added by QDarkStyle
            borderBottom='0px',
            # Padding for text inside tabs
            padding='4px 10px',
        )

        # -- Style for not selected tabs
        css['QTabBar::tab:!selected'].setValues(
            border='0px',
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_4,
            borderLeft=f'1px solid {SpyderPalette.COLOR_BACKGROUND_4}',
            borderRight=f'1px solid {SpyderPalette.SPECIAL_TABS_SEPARATOR}',
        )

        css['QTabBar::tab:!selected:hover'].setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_5,
            borderLeftColor=SpyderPalette.COLOR_BACKGROUND_5
        )

        # -- Style for the not selected tabs to the right and left of the
        # selected one.
        # Note: For some strange reason, Qt uses the `next-selected` state for
        # the left tab.
        css['QTabBar::tab:next-selected'].setValues(
            borderRightColor=SpyderPalette.COLOR_BACKGROUND_4,
        )

        css['QTabBar::tab:next-selected:hover'].setValues(
            borderRightColor=SpyderPalette.SPECIAL_TABS_SEPARATOR,
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_5
        )

        css['QTabBar::tab:previous-selected'].setValues(
            borderLeftColor=SpyderPalette.COLOR_BACKGROUND_4,
        )

        css['QTabBar::tab:previous-selected:hover'].setValues(
            borderLeftColor=SpyderPalette.SPECIAL_TABS_SEPARATOR,
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_5
        )

        # -- First and last tabs have rounded borders
        css['QTabBar::tab:first'].setValues(
            borderTopLeftRadius=SpyderPalette.SIZE_BORDER_RADIUS,
            borderBottomLeftRadius=SpyderPalette.SIZE_BORDER_RADIUS,
        )

        css['QTabBar::tab:last'].setValues(
            borderTopRightRadius=SpyderPalette.SIZE_BORDER_RADIUS,
            borderBottomRightRadius=SpyderPalette.SIZE_BORDER_RADIUS,
        )

        # -- Last tab doesn't need to show the separator
        css['QTabBar::tab:last:!selected'].setValues(
            borderRightColor=SpyderPalette.COLOR_BACKGROUND_4
        )

        css['QTabBar::tab:last:!selected:hover'].setValues(
            borderRightColor=SpyderPalette.COLOR_BACKGROUND_5,
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_5
        )

        # -- Set bottom margin for scroll buttons.
        css['QTabBar QToolButton'].setValues(
            marginBottom=f'{2 * margin_size}px',
        )


class PreferencesTabBarStyleSheet(SpecialTabBarStyleSheet, SpyderFontsMixin):
    """Style for tab bars in our Preferences dialog."""

    # This is necessary because this class needs to access fonts
    SET_STYLESHEET_AT_INIT = False

    def set_stylesheet(self):
        super().set_stylesheet()

        # Main constants
        css = self.get_stylesheet()
        font = self.get_font(SpyderFontType.Interface, font_size_delta=1)

        # Set font size to be one point bigger than the regular text.
        css.QTabBar.setValues(
            fontSize=f'{font.pointSize()}pt',
        )

        # Make scroll buttons a bit bigger on Windows and Mac (this has no
        # effect on Linux).
        if WIN or MAC:
            css['QTabBar QToolButton'].setValues(
                padding=f'{self.SCROLL_BUTTONS_PADDING - 1}px',
            )

        # Increase padding around text because we're using a larger font.
        css['QTabBar::tab'].setValues(
            padding='6px 10px',
        )

        # Remove border and add padding for content inside tabs
        css['QTabWidget::pane'].setValues(
            border='0px',
            padding=f'{AppStyle.InnerContentPadding}px',
        )


class HorizontalDockTabBarStyleSheet(SpecialTabBarStyleSheet):
    """Style for horizontal dockwidget tab bars."""

    def set_stylesheet(self):
        super().set_stylesheet()

        # Main constants
        css = self.get_stylesheet()
        margin_size = AppStyle.MarginSize

        # Tabs style
        css['QTabBar::tab'].setValues(
            # No margins to left/right but top/bottom to separate tabbar from
            # the dockwidget areas.
            # Notes:
            # * Top margin is half the one at the bottom so that we can show
            #   a bottom margin on dockwidgets that are not tabified.
            # * The other half is added through the _margin_bottom attribute of
            #   PluginMainWidget.
            margin=f'{margin_size}px 0px {2 * margin_size}px 0px',
            # Remove a colored border added by QDarkStyle
            borderTop='0px',
        )

        # Add margin to first and last tabs to avoid them touching the left and
        # right dockwidget areas, respectively.
        css['QTabBar::tab:first'].setValues(
            marginLeft=f'{2 * margin_size}px',
        )

        css['QTabBar::tab:last'].setValues(
            marginRight=f'{2 * margin_size}px',
        )

        # Make top and bottom margins for scroll buttons even.
        # This is necessary since the tabbar top margin is half the one at the
        # bottom (see the notes in the 'QTabBar::tab' style above).
        css['QTabBar QToolButton'].setValues(
            marginTop='0px',
            marginBottom=f'{margin_size}px',
        )


class VerticalDockTabBarStyleSheet(BaseDockTabBarStyleSheet):
    """Style for vertical dockwidget tab bars."""

    SCROLL_BUTTONS_BORDER_POS = 'bottom'

    def set_stylesheet(self):
        super().set_stylesheet()

        # -- Main constants
        css = self.get_stylesheet()
        margin_size = AppStyle.MarginSize

        # -- Basic style
        css['QTabBar::tab'].setValues(
            # No margins to top/bottom but left/right to separate tabbar from
            # the dockwidget areas
            margin=f'0px {2 * margin_size}px',
            # Border radius is added for specific tabs (see below)
            borderRadius='0px',
            # Remove colored borders added by QDarkStyle
            borderLeft='0px',
            borderRight='0px',
            # Padding for text inside tabs
            padding='10px 4px',
        )

        # -- Style for not selected tabs
        css['QTabBar::tab:!selected'].setValues(
            border='0px',
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_4,
            borderTop=f'1px solid {SpyderPalette.COLOR_BACKGROUND_4}',
            borderBottom=f'1px solid {SpyderPalette.SPECIAL_TABS_SEPARATOR}',
        )

        css['QTabBar::tab:!selected:hover'].setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_5,
            borderTopColor=SpyderPalette.COLOR_BACKGROUND_5,
        )

        # -- Style for the not selected tabs above and below the selected one.
        css['QTabBar::tab:next-selected'].setValues(
            borderBottomColor=SpyderPalette.COLOR_BACKGROUND_4,
        )

        css['QTabBar::tab:next-selected:hover'].setValues(
            borderBottomColor=SpyderPalette.SPECIAL_TABS_SEPARATOR,
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_5
        )

        css['QTabBar::tab:previous-selected'].setValues(
            borderTopColor=SpyderPalette.COLOR_BACKGROUND_4,
        )

        css['QTabBar::tab:previous-selected:hover'].setValues(
            borderTopColor=SpyderPalette.SPECIAL_TABS_SEPARATOR,
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_5
        )

        # -- First and last tabs have rounded borders.
        # Also, add margin to avoid them touch the top and bottom borders,
        # respectively.
        css['QTabBar::tab:first'].setValues(
            borderTopLeftRadius=SpyderPalette.SIZE_BORDER_RADIUS,
            borderTopRightRadius=SpyderPalette.SIZE_BORDER_RADIUS,
            marginTop=f'{2 * margin_size}px',
        )

        css['QTabBar::tab:last'].setValues(
            borderBottomLeftRadius=SpyderPalette.SIZE_BORDER_RADIUS,
            borderBottomRightRadius=SpyderPalette.SIZE_BORDER_RADIUS,
            marginBottom=f'{2 * margin_size}px',
        )

        # -- Last tab doesn't need to show the separator
        css['QTabBar::tab:last:!selected'].setValues(
            borderBottomColor=SpyderPalette.COLOR_BACKGROUND_4
        )

        css['QTabBar::tab:last:!selected:hover'].setValues(
            borderBottomColor=SpyderPalette.COLOR_BACKGROUND_5,
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_5
        )

        # -- Make style for scroll buttons match the horizontal one
        css['QTabBar QToolButton'].setValues(
            marginLeft=f'{margin_size}px',
            marginRight=f'{margin_size}px',
        )


PANES_TABBAR_STYLESHEET = PanesTabBarStyleSheet()
HORIZONTAL_DOCK_TABBAR_STYLESHEET = HorizontalDockTabBarStyleSheet()
VERTICAL_DOCK_TABBAR_STYLESHEET = VerticalDockTabBarStyleSheet()
PREFERENCES_TABBAR_STYLESHEET = PreferencesTabBarStyleSheet()


# =============================================================================
# ---- Style for special dialogs
# =============================================================================
class DialogStyle(SpyderFontsMixin):
    """Style constants for tour and about dialogs."""

    IconScaleFactor = 0.5
    BackgroundColor = SpyderPalette.COLOR_BACKGROUND_2
    BorderColor = SpyderPalette.COLOR_BACKGROUND_5

    @classproperty
    def _fs(cls):
        return cls.get_font(SpyderFontType.Interface).pointSize()

    @classproperty
    def TitleFontSize(cls):
        if WIN:
            return f"{cls._fs + 6}pt"
        elif MAC:
            return f"{cls._fs + 6}pt"
        else:
            return f"{cls._fs + 4}pt"

    @classproperty
    def ContentFontSize(cls):
        if WIN:
            return f"{cls._fs + 4}pt"
        elif MAC:
            return f"{cls._fs + 2}pt"
        else:
            return f"{cls._fs + 2}pt"

    @classproperty
    def ButtonsFontSize(cls):
        if WIN:
            return f"{cls._fs + 5}pt"
        elif MAC:
            return f"{cls._fs + 2}pt"
        else:
            return f"{cls._fs + 3}pt"

    @classproperty
    def ButtonsPadding(cls):
        if WIN:
            return f"{AppStyle.MarginSize + 1}px {5 * AppStyle.MarginSize}px"
        elif MAC:
            return f"{2 * AppStyle.MarginSize}px {4 * AppStyle.MarginSize}px"
        else:
            return f"{AppStyle.MarginSize + 1}px {AppStyle.MarginSize}px"



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for status.py
"""

# Test library imports
import pytest

# Third party imports
from qtpy.QtWidgets import QMainWindow

# Local imports
from spyder.plugins.editor.widgets.status import (CursorPositionStatus,
                                                  EncodingStatus, EOLStatus,
                                                  ReadWriteStatus, VCSStatus)


@pytest.fixture
def status_bar(qtbot):
    """Set up StatusBarWidget."""
    win = QMainWindow()
    win.setWindowTitle("Status widgets test")
    win.resize(900, 300)
    statusbar = win.statusBar()
    qtbot.addWidget(win)
    return (win, statusbar)


def test_status_bar(status_bar, qtbot):
    """Run StatusBarWidget."""
    win, statusbar = status_bar
    swidgets = []
    for klass in (ReadWriteStatus, EOLStatus, EncodingStatus,
                  CursorPositionStatus, VCSStatus):
        swidget = klass(win)
        swidgets.append(swidget)
    assert win
    assert len(swidgets) == 5


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Create Spyder installers using `constructor`.

It creates a `construct.yaml` file with the needed settings
and then runs `constructor`.

Some environment variables we use:

CONSTRUCTOR_TARGET_PLATFORM:
    conda-style platform (as in `platform` in `conda info -a` output)
CONSTRUCTOR_CONDA_EXE:
    when the target platform is not the same as the host, constructor
    needs a path to a conda-standalone (or micromamba) executable for
    that platform. needs to be provided in this env var in that case!
CONSTRUCTOR_SIGNING_CERTIFICATE:
    Path to PFX certificate to sign the EXE installer on Windows
"""

# Standard library imports
from argparse import ArgumentParser
from datetime import timedelta
from functools import partial
import json
from logging import getLogger
import os
from packaging.version import parse
from pathlib import Path
import platform
import re
import shutil
from subprocess import run
import sys
from textwrap import dedent, indent
from time import time
import zipfile

# Third-party imports
from ruamel.yaml import YAML
from setuptools_scm import get_version

# Local imports
from build_conda_pkgs import HERE, BUILD, RESOURCES, SPECS, h

DIST = HERE / "dist"

logger = getLogger('BuildInstallers')
logger.addHandler(h)
logger.setLevel('INFO')

APP = "Spyder"
SPYREPO = HERE.parent
WINDOWS = os.name == "nt"
MACOS = sys.platform == "darwin"
LINUX = sys.platform.startswith("linux")
CONDA_BLD_PATH = os.getenv("CONDA_BLD_PATH", "local")
PY_VER = "{v.major}.{v.minor}.{v.micro}".format(v=sys.version_info)
SPYVER = get_version(SPYREPO).split("+")[0]

if WINDOWS:
    OS = "Windows"
    TARGET_PLATFORM = "win-"
    INSTALL_CHOICES = ["exe"]
elif LINUX:
    OS = "Linux"
    TARGET_PLATFORM = "linux-"
    INSTALL_CHOICES = ["sh"]
elif MACOS:
    OS = "macOS"
    TARGET_PLATFORM = "osx-"
    INSTALL_CHOICES = ["pkg", "sh"]
else:
    raise RuntimeError(f"Unrecognized OS: {sys.platform}")

ARCH = (platform.machine() or "generic").lower().replace("amd64", "x86_64")
TARGET_PLATFORM = (TARGET_PLATFORM + ARCH).replace("x86_64", "64")
TARGET_PLATFORM = os.getenv("CONSTRUCTOR_TARGET_PLATFORM", TARGET_PLATFORM)

# ---- Parse arguments
p = ArgumentParser()
p.add_argument(
    "--no-local", action="store_true",
    help="Do not use local conda packages"
)
p.add_argument(
    "--debug", action="store_true",
    help="Do not delete build files"
)
p.add_argument(
    "--arch", action="store_true",
    help="Print machine architecture tag and exit.",
)
p.add_argument(
    "--ext", action="store_true",
    help="Print installer extension for this platform and exit.",
)
p.add_argument(
    "--artifact-name", action="store_true",
    help="Print computed artifact name and exit.",
)
p.add_argument(
    "--extra-specs", nargs="+", default=[],
    help="One or more extra conda specs to add to the installer.",
)
p.add_argument(
    "--licenses", action="store_true",
    help="Post-process licenses AFTER having built the installer. "
    "This must be run as a separate step.",
)
p.add_argument(
    "--images", action="store_true",
    help="Generate background images from the logo (test only)",
)
p.add_argument(
    "--cert-id", default=None,
    help="Apple Developer ID Application certificate common name."
)
p.add_argument(
    "--install-type", choices=INSTALL_CHOICES, default=INSTALL_CHOICES[0],
    help="Installer type."
)
p.add_argument(
    "--conda-lock", action="store_true",
    help="Create conda-lock file and exit."
)
p.add_argument(
    "--version", action="store_true",
    help="Print Spyder version and exit."
)
args = p.parse_args()

yaml = YAML()
yaml.indent(mapping=2, sequence=4, offset=2)
indent4 = partial(indent, prefix="    ")

base_specs = {
    "python": "=3.11.9",
    "conda": "=24.5.0",
    "menuinst": "=2.1.2",
}
rt_specs = {
    "python": f"={PY_VER}",
    "spyder": f"={SPYVER}",
    "cython": "",
    "matplotlib-base": "",
    "numpy": "",
    "openpyxl": "",
    "pandas": "",
    "scipy": "",
    "sympy": "",
}

if SPECS.exists():
    logger.info(f"Reading specs from {SPECS}...")
    _rt_specs = yaml.load(SPECS.read_text())
    rt_specs.update(_rt_specs)
else:
    logger.info(f"Did not read specs from {SPECS}")

for spec in args.extra_specs:
    k, *v = re.split('([<>=]+)[ ]*', spec)
    rt_specs[k] = "".join(v).strip()

PY_VER = parse(re.split('([<>=]+)[ ]*', rt_specs['python'])[-1])
SPYVER = parse(re.split('([<>=]+)[ ]*', rt_specs['spyder'])[-1])

BASE_LOCK_FILE = BUILD / f"conda-base-{TARGET_PLATFORM}.lock"
RT_LOCK_FILE = BUILD / f"conda-runtime-{TARGET_PLATFORM}.lock"
OUTPUT_FILE = DIST / f"{APP}-{OS}-{ARCH}.{args.install_type}"
INSTALLER_DEFAULT_PATH_STEM = f"{APP.lower()}-{SPYVER.major}"

WELCOME_IMG_WIN = BUILD / "welcome_img_win.png"
HEADER_IMG_WIN = BUILD / "header_img_win.png"
WELCOME_IMG_MAC = BUILD / "welcome_img_mac.png"
CONSTRUCTOR_FILE = BUILD / "construct.yaml"


def _create_conda_lock(env_type='base'):
    specs = base_specs
    lock_file = BASE_LOCK_FILE
    if env_type == "runtime":
        specs = rt_specs
        lock_file = RT_LOCK_FILE

    definitions = {
        "channels": [
            "conda-forge",
            "conda-forge/label/spyder_rc",
            "conda-forge/label/spyder_dev",
            "conda-forge/label/spyder_kernels_rc",
            # TODO: Add this channel when we create it
            # "conda-forge/label/spyder_kernels_dev",
            CONDA_BLD_PATH,
        ],
        "dependencies": [k + v for k, v in specs.items()],
        "platforms": [TARGET_PLATFORM]
    }

    if args.no_local:
        definitions['channels'].remove(CONDA_BLD_PATH)

    logger.info(f"Conda lock configuration ({env_type}):")
    if logger.getEffectiveLevel() <= 20:
        yaml.dump(definitions, sys.stdout)

    env_file = BUILD / f"{env_type}_env.yml"
    yaml.dump(definitions, env_file)

    env = os.environ.copy()
    env["CONDA_CHANNEL_PRIORITY"] = "flexible"

    cmd_args = [
        "conda-lock", "lock",
        "--kind", "explicit",
        "--file", str(env_file),
        "--filename-template", str(BUILD / f"conda-{env_type}-{{platform}}.lock")
        # Note conda-lock doesn't provide output file option, only template
    ]

    run(cmd_args, check=True, env=env)

    logger.info(f"Contents of {lock_file}:")
    if logger.getEffectiveLevel() <= 20:
        print(lock_file.read_text(), flush=True)

    # Write to dist directory also
    (DIST / lock_file.name).write_text(lock_file.read_text())


def _generate_background_images(installer_type):
    """This requires Pillow."""
    if installer_type == "sh":
        # shell installers are text-based, no graphics
        return

    from PIL import Image

    logo_path = SPYREPO / "img_src" / "spyder.png"
    logo = Image.open(logo_path, "r")

    if installer_type in ("exe", "all"):
        sidebar = Image.new("RGBA", (164, 314), (0, 0, 0, 0))
        sidebar.paste(logo.resize((101, 101)), (32, 180))
        output = WELCOME_IMG_WIN
        sidebar.save(output, format="png")

        banner = Image.new("RGBA", (150, 57), (0, 0, 0, 0))
        banner.paste(logo.resize((44, 44)), (8, 6))
        output = HEADER_IMG_WIN
        banner.save(output, format="png")

    if installer_type in ("pkg", "all"):
        _logo = Image.new("RGBA", logo.size, "WHITE")
        _logo.paste(logo, mask=logo)
        background = Image.new("RGBA", (1227, 600), (0, 0, 0, 0))
        background.paste(_logo.resize((148, 148)), (95, 418))
        output = WELCOME_IMG_MAC
        background.save(output, format="png")


def _get_condarc():
    contents = dedent(
        """
        channels:  #!final
          - conda-forge/label/spyder_rc
          - conda-forge/label/spyder_dev
          - conda-forge/label/spyder_kernels_rc
          # TODO: Add this channel when we create it
          # - conda-forge/label/spyder_kernels_dev
          - conda-forge
        repodata_fns:  #!final
          - repodata.json
        auto_update_conda: false  #!final
        notify_outdated_conda: false  #!final
        channel_priority: flexible  #!final
        env_prompt: '[spyder]({default_env}) '  #! final
        register_envs: false  #! final
        """
    )
    # the undocumented #!final comment is explained here
    # https://www.anaconda.com/blog/conda-configuration-engine-power-users
    file = BUILD / "condarc"
    file.write_text(contents)

    return str(file)


def _get_conda_bld_path_url():
    bld_path_url = "file://"
    if WINDOWS:
        bld_path_url += "/"
    bld_path_url += Path(os.getenv('CONDA_BLD_PATH')).as_posix()
    return bld_path_url


def _definitions():
    condarc = _get_condarc()
    definitions = {
        "name": APP,
        "company": "Spyder-IDE",
        "reverse_domain_identifier": "org.spyder-ide.Spyder",
        "version": str(SPYVER),
        "environment_file": str(BASE_LOCK_FILE),
        "installer_filename": OUTPUT_FILE.name,
        "installer_type": args.install_type,
        "initialize_by_default": False,
        "initialize_conda": False,
        "register_python": False,
        "register_envs": False,
        "extra_envs": {
            "spyder-runtime": {
                "environment_file": str(RT_LOCK_FILE),
            }
        },
        "channels_remap": [
            {
                "src": _get_conda_bld_path_url(),
                "dest": "https://conda.anaconda.org/conda-forge"
            }
        ]
    }

    definitions["license_file"] = str(RESOURCES / "bundle_license.rtf")
    if args.install_type == "sh":
        definitions["license_file"] = str(SPYREPO / "LICENSE.txt")

    if LINUX:
        definitions.update(
            {
                "default_prefix": os.path.join(
                    "$HOME", ".local", INSTALLER_DEFAULT_PATH_STEM
                ),
                "pre_install": str(RESOURCES / "pre-install.sh"),
                "post_install": str(RESOURCES / "post-install.sh"),
                "extra_files": [
                    {str(RESOURCES / "bundle_readme.md"): "README.txt"},
                    {condarc: ".condarc"},
                    {str(RESOURCES / "menuinst_cli.py"): "bin/menuinst_cli.py"},
                ],
            }
        )

    if MACOS:
        welcome_text_tmpl = \
            (RESOURCES / "osx_pkg_welcome.rtf.tmpl").read_text()
        welcome_file = BUILD / "osx_pkg_welcome.rtf"
        welcome_file.write_text(
            welcome_text_tmpl.replace("__VERSION__", str(SPYVER)))

        definitions.update(
            {
                "progress_notifications": True,
                "pre_install": str(RESOURCES / "pre-install.sh"),
                "post_install": str(RESOURCES / "post-install.sh"),
                "conclusion_text": "",
                "readme_text": "",
                # For sh installer
                "default_prefix": os.path.join(
                    "$HOME", "Library", INSTALLER_DEFAULT_PATH_STEM
                ),
                # For pkg installer
                "pkg_name": INSTALLER_DEFAULT_PATH_STEM,
                "default_location_pkg": "Library",
                "welcome_image": str(WELCOME_IMG_MAC),
                "welcome_file": str(welcome_file),
                "extra_files": [
                    {str(RESOURCES / "bundle_readme.md"): "README.txt"},
                    {condarc: ".condarc"},
                    {str(RESOURCES / "menuinst_cli.py"): "bin/menuinst_cli.py"},
                ],
            }
        )

        if args.cert_id:
            definitions["signing_identity_name"] = args.cert_id
            definitions["notarization_identity_name"] = args.cert_id

    if WINDOWS:
        definitions.update(
            {
                "uninstall_name": f"Spyder {SPYVER.major}",
                "welcome_image": str(WELCOME_IMG_WIN),
                "header_image": str(HEADER_IMG_WIN),
                "icon_image": str(SPYREPO / "img_src" / "spyder.ico"),
                "default_prefix": os.path.join(
                    "%LOCALAPPDATA%", INSTALLER_DEFAULT_PATH_STEM
                ),
                "default_prefix_domain_user": os.path.join(
                    "%LOCALAPPDATA%", INSTALLER_DEFAULT_PATH_STEM
                ),
                "default_prefix_all_users": os.path.join(
                    "%ALLUSERSPROFILE%", INSTALLER_DEFAULT_PATH_STEM
                ),
                "check_path_length": False,
                "pre_install": str(RESOURCES / "pre-install.bat"),
                "post_install": str(RESOURCES / "post-install.bat"),
                "extra_files": [
                    {str(RESOURCES / "bundle_readme.md"): "README.txt"},
                    {condarc: ".condarc"},
                    {str(RESOURCES / "menuinst_cli.py"): "Scripts/menuinst_cli.py"},
                ],
            }
        )

        signing_certificate = os.environ.get("CONSTRUCTOR_SIGNING_CERTIFICATE")
        if signing_certificate:
            definitions["signing_certificate"] = signing_certificate

    if definitions.get("welcome_image") or definitions.get("header_image"):
        _generate_background_images(definitions.get("installer_type", "all"))

    logger.info(f"Contents of {CONSTRUCTOR_FILE}:")
    if logger.getEffectiveLevel() <= 20:
        yaml.dump(definitions, sys.stdout)

    yaml.dump(definitions, CONSTRUCTOR_FILE)


def _constructor():
    """
    Create a temporary `construct.yaml` input file and
    run `constructor`.
    """
    constructor = shutil.which("constructor")
    if not constructor:
        raise RuntimeError("Constructor must be installed and in PATH.")

    _definitions()

    cmd_args = [
        constructor, "-v",
        "--output-dir", str(DIST),
        "--platform", TARGET_PLATFORM,
        str(CONSTRUCTOR_FILE.parent)
    ]
    if args.debug:
        cmd_args.append("--debug")
    conda_exe = os.getenv("CONSTRUCTOR_CONDA_EXE")
    if conda_exe:
        cmd_args.extend(["--conda-exe", conda_exe])

    env = os.environ.copy()
    env["CONDA_CHANNEL_PRIORITY"] = "flexible"

    logger.info("Command: " + " ".join(cmd_args))

    run(cmd_args, check=True, env=env)


def licenses():
    info_path = BUILD / "info.json"
    try:
        info = json.load(info_path)
    except FileNotFoundError:
        print(
            "!! Use `constructor --debug` to write info.json and get licenses",
            file=sys.stderr,
        )
        raise

    zipname = BUILD / f"licenses.{OS}-{ARCH}.zip"
    output_zip = zipfile.ZipFile(zipname, mode="w",
                                 compression=zipfile.ZIP_DEFLATED)
    output_zip.write(info_path)
    for package_id, license_info in info["_licenses"].items():
        package_name = package_id.split("::", 1)[1]
        for license_type, license_files in license_info.items():
            for i, license_file in enumerate(license_files, 1):
                arcname = (f"{package_name}.{license_type.replace(' ', '_')}"
                           f".{i}.txt")
                output_zip.write(license_file, arcname=arcname)
    output_zip.close()
    return zipname.resolve()


def main():
    t0 = time()
    try:
        DIST.mkdir(exist_ok=True)
        _create_conda_lock(env_type='base')
        assert BASE_LOCK_FILE.exists()
        _create_conda_lock(env_type='runtime')
        assert RT_LOCK_FILE.exists()
    finally:
        elapse = timedelta(seconds=int(time() - t0))
        logger.info(f"Build time: {elapse}")

    t0 = time()
    try:
        _constructor()
        assert OUTPUT_FILE.exists()
        logger.info(f"Created {OUTPUT_FILE}")
    finally:
        elapse = timedelta(seconds=int(time() - t0))
        logger.info(f"Build time: {elapse}")


if __name__ == "__main__":
    if args.arch:
        print(ARCH)
        sys.exit()
    if args.ext:
        print(args.install_type)
        sys.exit()
    if args.artifact_name:
        print(OUTPUT_FILE)
        sys.exit()
    if args.licenses:
        print(licenses())
        sys.exit()
    if args.images:
        _generate_background_images()
        sys.exit()
    if args.conda_lock:
        _create_conda_lock(env_type='base')
        _create_conda_lock(env_type='runtime')
        sys.exit()
    if args.version:
        print(SPYVER)
        sys.exit()

    main()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""
Tests for the spyder.config.base module.
"""

# Standard library imports
from importlib import reload
import os
import os.path as osp
from pathlib import Path
import shutil


# Third party imports
import pytest

# Local imports
import spyder.config.base
from spyder.config.utils import is_anaconda
from spyder.utils.conda import get_list_conda_envs


# ============================================================================
# ---- Tests
# ============================================================================
def test_is_stable_version():
    """Test that stable and non-stable versions are recognized correctly."""
    for stable_version in ['3.3.0', '2', ('0', '5')]:
        assert spyder.config.base.is_stable_version(stable_version)
    for not_stable_version in ['4.0.0b1', '3.3.2.dev0',
                               'beta', ('2', '0', 'alpha')]:
        assert not spyder.config.base.is_stable_version(not_stable_version)


@pytest.mark.parametrize('use_dev_config_dir', [True, False])
def test_get_conf_path(monkeypatch, use_dev_config_dir):
    """Test that the config dir path is set under dev and release builds."""
    monkeypatch.setenv('SPYDER_USE_DEV_CONFIG_DIR', str(use_dev_config_dir))
    monkeypatch.setenv('SPYDER_PYTEST', '')
    reload(spyder.config.base)
    conf_path = spyder.config.base.get_conf_path()
    assert conf_path
    assert ((osp.basename(conf_path).split('-')[-1] == 'dev')
            == use_dev_config_dir)
    assert osp.isdir(conf_path)
    monkeypatch.undo()
    reload(spyder.config.base)


@pytest.mark.skipif(
    not spyder.config.base.running_in_ci(), reason="Only works on CIs"
)
@pytest.mark.skipif(not is_anaconda(), reason='Only works with Anaconda')
def test_is_conda_based_app():
    """Test that is_conda_based_app is working as expected."""
    # Get conda env to use
    pyexec = get_list_conda_envs()['Conda: jedi-test-env'][0]

    # Get env's root
    env_root = (
        Path(pyexec).parents[0] if os.name == "nt" else Path(pyexec).parents[1]
    )

    # Create dir and file necessary to detect the app
    menu_dir = env_root / "Menu"
    menu_dir.mkdir()
    (menu_dir / "conda-based-app").touch()

    # Check the env is detected as belonging to the app
    assert spyder.config.base.is_conda_based_app(pyexec=pyexec)

    # Remove added dir
    shutil.rmtree(menu_dir, ignore_errors=True)


if __name__ == '__main__':
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

# Standard library imports
import os
import sys
import tempfile

# Third party imports
import pytest
from qtpy.QtGui import QTextOption

# Local imports
from spyder.config.base import get_conf_path, running_in_ci
from spyder.config.manager import CONF
from spyder.plugins.history import plugin as history


# =============================================================================
# Utillity Functions
# =============================================================================
def create_file(name, content):
    path = os.path.join(tempfile.gettempdir(), name)
    with open(path, 'w') as fh:
        fh.write(content)

    return path


# =============================================================================
# Qt Test Fixtures
# =============================================================================
@pytest.fixture
def historylog(qtbot):
    """
    Return a fixture for base history log, which is a plugin widget.
    """
    historylog = history.HistoryLog(None, configuration=CONF)
    historylog.close = lambda: True
    qtbot.addWidget(historylog.get_widget())
    historylog.get_widget().show()
    yield historylog
    historylog.on_close()
    historylog.close()


# =============================================================================
# Tests
# =============================================================================
def test_init(historylog):
    """
    Test HistoryLog

    Test that the initialization created the expected instance variables
    and widgets for a new HistoryLog instance.
    """
    hl = historylog
    assert len(hl.get_widget().editors) == 1
    assert len(hl.get_widget().filenames) == 1
    assert len(hl.get_actions()) == 7


@pytest.mark.order(1)
@pytest.mark.skipif(
    sys.platform == "darwin" and running_in_ci(),
    reason="Fails on Mac when running on CI "
)
def test_add_history(historylog):
    """
    Test the add_history method.

    Test adding a history file to the history log widget and the
    code editor properties that are enabled/disabled.
    """
    hl = historylog
    hw = historylog.get_widget()

    # No editors yet.
    assert len(hw.editors) == 1

    # Add one file.
    name1 = 'test_history.py'
    text1 = 'a = 5\nb= 10\na + b\n'
    path1 = create_file(name1, text1)
    tab1 = os.path.basename(path1)
    hw.set_conf('line_numbers', False)
    hw.set_conf('wrap', False)
    hl.add_history(path1)

    # Check tab and editor were created correctly.
    assert len(hw.editors) == 2
    assert hw.filenames == [get_conf_path('history.py'), path1]
    assert hw.tabwidget.currentIndex() == 1
    assert not hw.editors[0].linenumberarea.isVisible()
    assert hw.editors[0].wordWrapMode() == QTextOption.NoWrap
    assert hw.tabwidget.tabText(1) == tab1
    assert hw.tabwidget.tabToolTip(1) == path1

    # Try to add same file -- does not process filename again, so
    # linenumbers and wrap doesn't change.
    hw.add_history(path1)
    assert hw.tabwidget.currentIndex() == 1
    # assert not hw.editors[0].linenumberarea.isVisible()

    # Add another file.
    name2 = 'history2.js'
    text2 = 'random text\nspam line\n\n\n\n'
    path2 = create_file(name2, text2)
    tab2 = os.path.basename(path2)
    hw.set_conf('line_numbers', True)
    hw.set_conf('wrap', True)
    hw.add_history(path2)

    # Check second tab and editor were created correctly.
    assert len(hw.editors) == 3
    assert hw.filenames == [get_conf_path('history.py'), path1, path2]
    assert hw.tabwidget.currentIndex() == 2
    assert hw.editors[2].wordWrapMode() == (
        QTextOption.WrapAtWordBoundaryOrAnywhere)
    assert hw.editors[2].linenumberarea.isVisible()
    assert hw.tabwidget.tabText(2) == tab2
    assert hw.tabwidget.tabToolTip(2) == path2

    # Check differences between tabs based on setup.
    assert hw.editors[1].supported_language
    assert hw.editors[1].isReadOnly()
    assert not hw.editors[1].isVisible()
    assert hw.editors[1].toPlainText() == text1

    assert not hw.editors[2].supported_language
    assert hw.editors[2].isReadOnly()
    assert hw.editors[2].isVisible()
    assert hw.editors[2].toPlainText() == text2


def test_append_to_history(qtbot, historylog):
    """
    Test the append_to_history method.

    Test adding text to a history file.  Also test the go_to_eof config
    option for positioning the cursor.
    """
    hw = historylog.get_widget()

    # Toggle to move to the end of the file after appending.
    hw.set_conf('go_to_eof', True)

    # Force cursor to the beginning of the file.
    text1 = 'import re\n'
    path1 = create_file('test_history.py', text1)
    hw.add_history(path1)
    hw.editors[1].set_cursor_position('sof')
    hw.append_to_history(path1, 'foo = "bar"\n')
    assert hw.editors[1].toPlainText() == text1 + 'foo = "bar"\n'
    assert hw.tabwidget.currentIndex() == 1

    # Cursor moved to end.
    assert hw.editors[1].is_cursor_at_end()
    assert not hw.editors[1].linenumberarea.isVisible()

    # Toggle to not move cursor after appending.
    hw.set_conf('go_to_eof', False)

    # Force cursor to the beginning of the file.
    hw.editors[1].set_cursor_position('sof')
    hw.append_to_history(path1, 'a = r"[a-z]"\n')
    assert hw.editors[1].toPlainText() == ('import re\n'
                                           'foo = "bar"\n'
                                           'a = r"[a-z]"\n')

    # Cursor not at end.
    assert not hw.editors[1].is_cursor_at_end()


def test_toggle_wrap_mode(historylog):
    """
    Test the toggle_wrap_mode method.

    Toggle the 'Wrap lines' config action.
    """
    hw = historylog.get_widget()
    path = create_file('test.py', 'a = 1')
    hw.add_history(path)

    # Starts with wrap mode off.
    hw.set_conf('wrap', False)
    assert hw.editors[1].wordWrapMode() == QTextOption.NoWrap
    assert not hw.get_conf('wrap')

    # Toggles wrap mode on.
    hw.set_conf('wrap', True)
    assert hw.editors[1].wordWrapMode() == (
        QTextOption.WrapAtWordBoundaryOrAnywhere)
    assert hw.get_conf('wrap')

    # Toggles wrap mode off.
    hw.set_conf('wrap', False)
    assert hw.editors[1].wordWrapMode() == QTextOption.NoWrap
    assert not hw.get_conf('wrap')


def test_toggle_line_numbers(historylog):
    """
    Test toggle_line_numbers method.

    Toggle the 'Show line numbers' config action.
    """
    hw = historylog.get_widget()
    path = create_file('test.py', 'a = 1')
    hw.add_history(path)

    # Starts without line numbers.
    hw.set_conf('line_numbers', False)
    assert not hw.editors[1].linenumberarea.isVisible()
    assert not hw.get_conf('line_numbers')

    # Toggles line numbers on.
    hw.set_conf('line_numbers', True)
    assert hw.editors[1].linenumberarea.isVisible()
    assert hw.get_conf('line_numbers')

    # Toggles line numbers off.
    hw.set_conf('line_numbers', False)
    assert not hw.editors[1].linenumberarea.isVisible()
    assert not hw.get_conf('line_numbers')


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-

# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Standard lib imports
import json
import os.path as osp

# Third-party imports
import pytest
import random

# Local imports
from spyder.config.snippets import SNIPPETS
from spyder.plugins.completion.api import (
    CompletionRequestTypes, CompletionItemKind)

PY_SNIPPETS = SNIPPETS['python']


@pytest.mark.order(3)
@pytest.mark.parametrize('trigger', list(PY_SNIPPETS.keys()))
def test_snippet_completions(qtbot_module, snippets_completions, trigger):
    snippets, completions = snippets_completions
    end_trim = random.randrange(1, len(trigger))
    descriptions = PY_SNIPPETS[trigger]
    expected_snippets = []
    for description in descriptions:
        snippet_info = descriptions[description]
        text = snippet_info['text']
        remove_trigger = snippet_info['remove_trigger']
        expected_snippets.append({
            'kind': CompletionItemKind.SNIPPET,
            'insertText': text,
            'label': f'{trigger} ({description})',
            'sortText': f'zzz{trigger}',
            'filterText': trigger,
            'documentation': '',
            'provider': 'Snippets',
            'remove_trigger': remove_trigger
        })

    trigger_text = trigger[:end_trim]
    snippets_request = {
        'file': '',
        'current_word': trigger_text
    }

    with qtbot_module.waitSignal(completions.sig_recv_snippets,
                                 timeout=3000) as blocker:
        snippets.send_request(
            'python',
            CompletionRequestTypes.DOCUMENT_COMPLETION,
            snippets_request
        )

    resp_snippets = blocker.args[0]
    resp_snippets = [x for x in resp_snippets if x['filterText'] == trigger]
    resp_snippets = sorted(resp_snippets, key=lambda x: x['label'])
    expected_snippets = sorted(expected_snippets, key=lambda x: x['label'])
    assert resp_snippets == expected_snippets



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

# Standard library imports
import os.path as osp
import sys
from unittest.mock import MagicMock

# Third party imports
from qtpy import QT_VERSION
from qtpy.QtCore import Qt, QEvent
from qtpy.QtGui import QTextCursor, QMouseEvent
from qtpy.QtWidgets import QApplication, QMainWindow, QTextEdit
import pytest

# Local imports
from spyder.config.base import running_in_ci
from spyder.plugins.preferences.tests.conftest import config_dialog
from spyder.plugins.shortcuts.plugin import Shortcuts
from spyder.widgets.mixins import TIP_PARAMETER_HIGHLIGHT_COLOR


HERE = osp.dirname(osp.abspath(__file__))
ASSETS = osp.join(HERE, 'assets')


class MainWindow(QMainWindow):

    _cli_options = MagicMock()

    def get_plugin(self, name, error=True):
        return MagicMock()


def test_editor_upper_to_lower(codeeditor):
    widget = codeeditor
    text = 'UPPERCASE'
    widget.set_text(text)
    cursor = widget.textCursor()
    cursor.movePosition(QTextCursor.NextCharacter,
                        QTextCursor.KeepAnchor)
    widget.setTextCursor(cursor)
    widget.transform_to_lowercase()
    new_text = widget.get_text('sof', 'eof')
    assert text != new_text


def test_editor_lower_to_upper(codeeditor):
    widget = codeeditor
    text = 'uppercase'
    widget.set_text(text)
    cursor = widget.textCursor()
    cursor.movePosition(QTextCursor.NextCharacter,
                        QTextCursor.KeepAnchor)
    widget.setTextCursor(cursor)
    widget.transform_to_uppercase()
    new_text = widget.get_text('sof', 'eof')
    assert text != new_text


@pytest.mark.parametrize(
    "input_text, expected_text, keys, strip_all",
    [
        ("for i in range(2): ",
         "for i in range(2): \n    \n     \n    ",
         [Qt.Key_Enter, Qt.Key_Enter, ' ', Qt.Key_Enter],
         False),
        ('for i in range(2): ',
         'for i in range(2):\n\n    ',
         [Qt.Key_Enter, Qt.Key_Enter],
         True),
        ('myvar = 2 ',
         'myvar = 2\n',
         [Qt.Key_Enter],
         True),
        ('somecode = 1\nmyvar = 2 \nmyvar = 3',
         'somecode = 1\nmyvar = 2 \nmyvar = 3',
         [' ', Qt.Key_Up, Qt.Key_Up],
         True),
        ('somecode = 1\nmyvar = 2 ',
         'somecode = 1\nmyvar = 2 ',
         [Qt.Key_Left],
         True),
        ('"""This is a string with important spaces\n    ',
         '"""This is a string with important spaces\n    \n',
         [Qt.Key_Enter],
         True),
        ('"""string   ',
         '"""string   \n',
         [Qt.Key_Enter],
         True),
        ('somecode = 1\nmyvar = 2',
         'somecode = 1\nmyvar = 2',
         [' ', (Qt.LeftButton, 0)],
         True),
        ('somecode = 1\nmyvar = 2',
         'somecode = 1\nmyvar = 2 ',
         [' ', (Qt.LeftButton, 23)],
         True),
        ('a=1\na=2 \na=3',
         'a=1\na=2 \na=3',
         [(Qt.LeftButton, 6), Qt.Key_Up],
         True),
        ('def fun():\n    """fun',
         'def fun():\n    """fun\n\n    ',
         [Qt.Key_Enter, Qt.Key_Enter],
         True),
        ('def fun():\n    """fun',
         'def fun():\n    """fun\n    \n    ',
         [Qt.Key_Enter, Qt.Key_Enter],
         False),
        ("('🚫')",
         "('🚫')\n",
         [Qt.Key_Enter],
         True),
        ("def fun():",
         "def fun():\n\n    ",
         [Qt.Key_Enter, Qt.Key_Enter],
         True),
        ("def fun():",
         "def fun():\n\n\n",
         [Qt.Key_Enter, Qt.Key_Enter, Qt.Key_Enter],
         True),
        ("def fun():\n    i = 0\n# no indent",
         "def fun():\n    i = 0\n# no indent\n",
         [Qt.Key_Enter],
         True),
        ("if a:\n    def b():\n        i = 1",
         "if a:\n    def b():\n        i = 1\n\n    ",
         [Qt.Key_Enter, Qt.Key_Enter, Qt.Key_Backspace],
         True),
    ])
def test_editor_rstrip_keypress(codeeditor, qtbot, input_text, expected_text,
                                keys, strip_all):
    """
    Test that whitespace is removed when leaving a line.
    """
    widget = codeeditor
    widget.strip_trailing_spaces_on_modify = strip_all
    widget.set_text(input_text)
    cursor = widget.textCursor()
    cursor.movePosition(QTextCursor.End)
    widget.setTextCursor(cursor)
    for key in keys:
        if isinstance(key, tuple):
            # Mouse event
            button, position = key
            cursor = widget.textCursor()
            cursor.setPosition(position)
            xypos = widget.cursorRect(cursor).center()
            widget.mousePressEvent(QMouseEvent(
                    QEvent.MouseButtonPress, xypos,
                    button, button,
                    Qt.NoModifier))
        else:
            qtbot.keyPress(widget, key)
    assert widget.toPlainText() == expected_text


@pytest.mark.parametrize(
    "input_text, expected_state", [
        ("'string ", [True, False]),
        ('"string ', [True, False]),
        ("'string \\", [True, True]),
        ('"string \\', [True, True]),
        ("'string \\ ", [True, False]),
        ('"string \\ ', [True, False]),
        ("'string ' ", [False, False]),
        ('"string " ', [False, False]),
        ("'string \"", [True, False]),
        ('"string \'', [True, False]),
        ("'string \" ", [True, False]),
        ('"string \' ', [True, False]),
        ("'''string ", [True, True]),
        ('"""string ', [True, True]),
        ("'''string \\", [True, True]),
        ('"""string \\', [True, True]),
        ("'''string \\ ", [True, True]),
        ('"""string \\ ', [True, True]),
        ("'''string ''' ", [False, False]),
        ('"""string """ ', [False, False]),
        ("'''string \"\"\"", [True, True]),
        ('"""string \'\'\'', [True, True]),
        ("'''string \"\"\" ", [True, True]),
        ('"""string \'\'\' ', [True, True]),
    ])
def test_in_string(codeeditor, input_text, expected_state):
    """
    Test that in_string works correctly.
    """
    widget = codeeditor
    widget.set_text(input_text + '\n  ')
    cursor = widget.textCursor()

    for blanks_enabled in [True, False]:
        widget.set_blanks_enabled(blanks_enabled)

        cursor.setPosition(len(input_text))
        assert cursor.position() == len(input_text)
        assert widget.in_string(cursor) == expected_state[0]

        cursor.setPosition(len(input_text) + 3)
        assert widget.in_string(cursor) == expected_state[1]


def test_comment(codeeditor):
    """
    Test that in_string works correctly.
    """
    widget = codeeditor
    widget.set_text("import numpy")
    cursor = widget.textCursor()
    cursor.setPosition(8)
    cursor.setPosition(11, QTextCursor.KeepAnchor)
    widget.setTextCursor(cursor)
    widget.toggle_comment()
    assert widget.toPlainText() == "# import numpy"
    widget.toggle_comment()
    assert widget.toPlainText() == "import numpy"


def test_undo_return(codeeditor, qtbot):
    """Test that we can undo a return."""
    editor = codeeditor
    text = "if True:\n    0"
    returned_text = "if True:\n    0\n    "
    editor.set_text(text)
    cursor = editor.textCursor()
    cursor.setPosition(14)
    editor.setTextCursor(cursor)
    qtbot.keyPress(editor, Qt.Key_Return)
    assert editor.toPlainText() == returned_text
    qtbot.keyPress(editor, "z", modifier=Qt.ControlModifier)
    assert editor.toPlainText() == text


def test_brace_match(codeeditor):
    """Tests for the highlighting of matching parenthesis, braces and brackets.

    Specifically provides regression tests for issues
     * spyder-ide/spyder#2965
     * spyder-ide/spyder#9179
     * spyder-ide/spyder#14374

    If this test fails the best way to investigate is probably to open
    assets/braces.py in Spyder, step through the file and visually
    observe brace matching.

    Some caveats for brace matching can be found in pull request
    spyder-ide/spyder#14376

    The functions being tested are essentially:
     * TextEditBaseWidget.find_brace_match
     * CodeEditor.in_comment
     * CodeEditor.in_string
    """
    # Create editor with contents loaded from assets/brackets.py
    editor = codeeditor
    with open(osp.join(ASSETS, 'braces.py'), 'r') as file:
        editor.set_text(file.read())

    # Each element of *positions* is a two element list:
    #  [position, expected]
    # Here *position* is the position at which to place the cursor and
    # *expected* is what editor.bracepos should be at that location if
    # the brace matching works correctly. Specifically if at *position* ...
    # a) ... there is no brace, then *expected* should be None.
    # b) ... there is an unmatched brace, then *expected* should be a
    #        1-tuple containing position-1
    # c) ... there is a matched brace then *expected* should be a
    #        2-tuple with the first element being position-1 and the
    #        second element being the position of the mathing brace.
    # At the end of each row, a comment has been added that attempts to
    # illustrate in what part of 'braces.py' the cursor is placed in
    # that test case.
    positions = [
        [0, None],       # b
        [5,  (4, 55)],   # b = [
        [56, (55,  4)],  # ]
        [7,  (6, 12)],   # [x
        [13, (12,  6)],  # x*2]
        [29, (28, 54)],  # [1
        [55, (54, 28)],  # ]
        [32, (31, 35)],  # [2
        [36, (35, 31)],  # 3]
        [38, (37, 53)],  # [4
        [54, (53, 37)],  # ]
        [41, (40, 42)],  # [5
        [42, None],      # 5
        [43, (42, 40)],  # 5]
        [47, (46, 52)],  # [7
        [53, (52, 46)],  # 8]
        [63, (62, 143)],  # a = [
        [144, (143, 62)],  # ]
        [69, (68, )],    # """(
        [70, (69, 78)],  # (
        [71, (70, 77)],  # (
        [72, (71, 76)],  # (
        [73, (72, 75)],  # (
        [74, (73, 74)],  # (
        [75, (74, 73)],  # )
        [76, (75, 72)],  # )
        [77, (76, 71)],  # )
        [78, (77, 70)],  # )
        [79, (78, 69)],  # )
        [82, (81, 88)],  # [
        [83, (82, 87)],  # [
        [84, (83, 86)],  # [
        [85, (84, 85)],  # [
        [86, (85, 84)],  # ]
        [87, (86, 83)],  # ]
        [88, (87, 82)],  # ]
        [89, (88, 81)],  # ]
        [90, (89, )],    # ]"""
        [99, (98, )],    # 'x)'
        [105, (104, )],  # 'b('
        [111, (110, )],  # # )
        [112, (111, 128)],  # {[
        [129, (128, 111)],  # ]}
        [113, (112, 127)],  # [(
        [128, (127, 112)],  # )]
        [114, (113, 126)],  # (
        [127, (126, 113)],  # )
    ]
    cursor = editor.textCursor()
    for position, expected in positions:
        cursor.setPosition(position)
        editor.setTextCursor(cursor)
        assert editor.bracepos == expected


def test_editor_backspace_char(codeeditor, qtbot):
    """Regression test for issue spyder-ide/spyder#12663."""
    editor = codeeditor
    text = "0123456789\nabcdefghij\n9876543210\njihgfedcba\n"
    editor.set_text(text)
    expected_column = 7
    cursor = editor.textCursor()
    cursor.setPosition(expected_column)
    editor.setTextCursor(cursor)
    for line in range(3):
        qtbot.keyPress(editor, Qt.Key_Backspace)
        expected_column -= 1
        assert editor.textCursor().columnNumber() == expected_column
        qtbot.keyPress(editor, Qt.Key_Down)
        assert editor.textCursor().columnNumber() == expected_column

    for line in range(3):
        qtbot.keyPress(editor, Qt.Key_Backspace)
        expected_column -= 1
        assert editor.textCursor().columnNumber() == expected_column
        qtbot.keyPress(editor, Qt.Key_Up)
        assert editor.textCursor().columnNumber() == expected_column


def test_editor_backspace_selection(codeeditor, qtbot):
    """Regression test for issue spyder-ide/spyder#12663."""
    editor = codeeditor
    text = "0123456789\nabcdefghij\n9876543210\njihgfedcba\n"
    editor.set_text(text)
    expected_column = 5
    cursor = editor.textCursor()
    cursor.setPosition(expected_column)
    editor.setTextCursor(cursor)

    # This first subtest does not trigger the original bug
    for press in range(3):
        qtbot.keyPress(editor, Qt.Key_Left, Qt.ShiftModifier)
    expected_column -= 3
    qtbot.keyPress(editor, Qt.Key_Backspace)
    assert editor.textCursor().columnNumber() == expected_column
    qtbot.keyPress(editor, Qt.Key_Down)
    assert editor.textCursor().columnNumber() == expected_column

    # However, this second subtest does trigger the original bug
    for press in range(3):
        qtbot.keyPress(editor, Qt.Key_Right, Qt.ShiftModifier)
    qtbot.keyPress(editor, Qt.Key_Backspace)
    assert editor.textCursor().columnNumber() == expected_column
    qtbot.keyPress(editor, Qt.Key_Down)
    assert editor.textCursor().columnNumber() == expected_column


def test_editor_delete_char(codeeditor, qtbot):
    """Regression test for issue spyder-ide/spyder#12663."""
    editor = codeeditor
    text = "0123456789\nabcdefghij\n9876543210\njihgfedcba\n"
    editor.set_text(text)
    expected_column = 2
    cursor = editor.textCursor()
    cursor.setPosition(expected_column)
    editor.setTextCursor(cursor)
    for line in range(3):
        qtbot.keyPress(editor, Qt.Key_Delete)
        assert editor.textCursor().columnNumber() == expected_column
        qtbot.keyPress(editor, Qt.Key_Down)
        assert editor.textCursor().columnNumber() == expected_column

    for line in range(3):
        qtbot.keyPress(editor, Qt.Key_Delete)
        assert editor.textCursor().columnNumber() == expected_column
        qtbot.keyPress(editor, Qt.Key_Up)
        assert editor.textCursor().columnNumber() == expected_column


# Fails in CI Linux tests, but not necessarily on all Linux installations
@pytest.mark.skipif(sys.platform.startswith('linux'), reason='Fail on Linux')
def test_editor_delete_selection(codeeditor, qtbot):
    """Regression test for issue spyder-ide/spyder#12663."""
    editor = codeeditor
    text = "0123456789\nabcdefghij\n9876543210\njihgfedcba\n"
    editor.set_text(text)
    expected_column = 5
    cursor = editor.textCursor()
    cursor.setPosition(expected_column)
    editor.setTextCursor(cursor)

    # This first subtest does not trigger the original bug
    for press in range(3):
        qtbot.keyPress(editor, Qt.Key_Left, Qt.ShiftModifier)
    expected_column -= 3
    qtbot.keyPress(editor, Qt.Key_Delete)
    assert editor.textCursor().columnNumber() == expected_column
    qtbot.keyPress(editor, Qt.Key_Down)
    assert editor.textCursor().columnNumber() == expected_column

    # However, this second subtest does trigger the original bug
    for press in range(3):
        qtbot.keyPress(editor, Qt.Key_Right, Qt.ShiftModifier)
    qtbot.keyPress(editor, Qt.Key_Delete)
    assert editor.textCursor().columnNumber() == expected_column
    qtbot.keyPress(editor, Qt.Key_Up)
    assert editor.textCursor().columnNumber() == expected_column


@pytest.mark.skipif(QT_VERSION.startswith('5.15'),
                    reason='Fixed on Qt 5.15')
def test_qtbug35861(qtbot):
    """This test will detect if upstream QTBUG-35861 is fixed.
    If that happens, then the workarounds for spyder-ide/spyder#12663
    can be removed. Such a fix would probably only happen in the most
    recent Qt version however...
    See also https://bugreports.qt.io/browse/QTBUG-35861
    """
    widget = QTextEdit(None)
    qtbot.addWidget(widget)
    widget.show()

    cursor = widget.textCursor()
    cursor.setPosition(0)
    # Build the text from a single character since a non-fixed width
    # font is used by default.
    cursor.insertText("0000000000\n"*5)

    expected_column = 5
    cursor.setPosition(expected_column)
    widget.setTextCursor(cursor)

    assert widget.textCursor().columnNumber() == expected_column
    for line in range(4):
        qtbot.keyClick(widget, Qt.Key_Backspace)
        assert widget.textCursor().columnNumber() == (expected_column - 1)
        qtbot.keyClick(widget, Qt.Key_Down)
        assert widget.textCursor().columnNumber() == expected_column

    for line in range(4):
        qtbot.keyClick(widget, Qt.Key_Backspace)
        assert widget.textCursor().columnNumber() == (expected_column - 1)
        qtbot.keyClick(widget, Qt.Key_Up)
        assert widget.textCursor().columnNumber() == expected_column


@pytest.mark.parametrize(
    "text",
    [
        "def foo(x):\n    return x\n",      # LF
        "def foo(x):\r\n    return x\r\n",  # CRLF
        "def foo(x):\r    return x\r"       # CR
    ]
)
def test_get_text_with_eol(codeeditor, text):
    """
    Test that get_text_with_eol returns the right text with the most
    common line endings.
    """
    editor = codeeditor
    editor.set_text(text)
    assert editor.get_text_with_eol() == text


def test_format_signature(codeeditor):
    """Test signature format method."""
    signature = """
    concatenate((a1, a2, a...), [b1, b2, b...], axis={}, index=[],
                *args, **kargs)"""
    editor = codeeditor
    color = TIP_PARAMETER_HIGHLIGHT_COLOR

    format_signature = editor._format_signature(signature, parameter="(a1")

    assert f"color:{color}'><b>a1</b></span>" in format_signature

    format_signature = editor._format_signature(signature, parameter="a2")
    assert f"color:{color}'><b>a2</b></span>" in format_signature

    format_signature = editor._format_signature(signature, parameter="a...")
    print(format_signature)
    assert f"color:{color}'><b>a...</b></span>" in format_signature

    format_signature = editor._format_signature(signature, parameter="[b1")
    assert f"color:{color}'><b>b1</b></span>" in format_signature

    format_signature = editor._format_signature(signature, parameter="b2")
    assert f"color:{color}'><b>b2</b></span>" in format_signature

    format_signature = editor._format_signature(signature, parameter="b...")
    assert f"color:{color}'><b>b...</b></span>" in format_signature

    format_signature = editor._format_signature(signature, parameter="axis")
    assert f"color:{color}'><b>axis</b></span>" in format_signature

    format_signature = editor._format_signature(signature, parameter="index")
    assert f"color:{color}'><b>index</b></span>" in format_signature

    format_signature = editor._format_signature(signature, parameter="*args")
    assert f"color:{color}'><b>*args</b></span>" in format_signature

    format_signature = editor._format_signature(signature, parameter="**kargs")
    assert f"color:{color}'><b>**kargs</b></span>" in format_signature


def test_delete(codeeditor):
    """Test CodeEditor.delete()."""
    editor = codeeditor
    text = ('def f1(a, b):\n')
    editor.set_text(text)

    # Move to start and delete next character without selection.
    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.Start)
    editor.setTextCursor(cursor)
    editor.delete()
    assert editor.get_text_line(0) == 'ef f1(a, b):'

    # Delete selection.
    cursor = editor.textCursor()
    cursor.select(QTextCursor.WordUnderCursor)
    editor.setTextCursor(cursor)
    editor.delete()
    assert editor.get_text_line(0) == ' f1(a, b):'

    # Move to end of document - nothing to delete after cursor.
    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.End)
    editor.setTextCursor(cursor)
    editor.delete()
    assert editor.get_text_line(0) == ' f1(a, b):'


def test_paste_files(codeeditor, copy_files_clipboard):
    """Test pasting files/folders into the editor."""
    editor = codeeditor
    file_paths = copy_files_clipboard
    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.Start)
    editor.setTextCursor(cursor)
    editor.paste()
    editor.selectAll()
    text = editor.toPlainText()
    path_list_in_editor = [path.strip(',"') for path in text.splitlines()]
    assert len(file_paths) == len(path_list_in_editor)
    for path, expected_path in zip(path_list_in_editor, file_paths):
        assert osp.normpath(path) == osp.normpath(expected_path)


@pytest.mark.parametrize('line_ending_char', ['\n', '\r\n', '\r'])
@pytest.mark.parametrize('text', ['def fun(a, b):\n\treturn a + b',
                                  'https://www.spyder-ide.org'])
def test_paste_text(codeeditor, text, line_ending_char):
    """Test pasting text into the editor."""
    editor = codeeditor
    text = text.replace(osp.os.linesep, line_ending_char)
    cb = QApplication.clipboard()
    cb.setText(text, mode=cb.Clipboard)
    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.Start)
    editor.setTextCursor(cursor)
    editor.paste()
    for line_no, txt in enumerate(text.splitlines()):
        assert editor.get_text_line(line_no) == txt


def test_cell_highlight(codeeditor, qtbot):
    """Test cells are properly highlighted."""
    editor = codeeditor
    text = ('\n\n\n#%%\n\n\n')
    editor.set_text(text)

    # Set cursor to start of file
    cursor = editor.textCursor()
    cursor.setPosition(0)
    editor.setTextCursor(cursor)
    assert editor.current_cell[0].selectionStart() == 0
    assert editor.current_cell[0].selectionEnd() == 3

    # Set cursor to start second cell
    cursor = editor.textCursor()
    cursor.setPosition(6)
    editor.setTextCursor(cursor)
    assert editor.current_cell[0].selectionStart() == 3
    assert editor.current_cell[0].selectionEnd() == 9

    # Delete cell
    qtbot.keyPress(editor, Qt.Key_Backspace)
    assert editor.current_cell[0].selectionStart() == 0
    assert editor.current_cell[0].selectionEnd() == 8

    # Create cell
    qtbot.keyPress(editor, "%")
    assert editor.current_cell[0].selectionStart() == 3
    assert editor.current_cell[0].selectionEnd() == 9

    # Test delete
    cursor = editor.textCursor()
    cursor.setPosition(5)
    editor.setTextCursor(cursor)
    qtbot.keyPress(editor, Qt.Key_Delete)
    assert editor.current_cell[0].selectionStart() == 0
    assert editor.current_cell[0].selectionEnd() == 8

    # Test undo
    editor.undo()
    assert editor.current_cell[0].selectionStart() == 3
    assert editor.current_cell[0].selectionEnd() == 9
    editor.redo()
    assert editor.current_cell[0].selectionStart() == 0
    assert editor.current_cell[0].selectionEnd() == 8


@pytest.mark.parametrize(
    'config_dialog',
    # [[MainWindowMock, [ConfigPlugins], [Plugins]]]
    [[MainWindow, [], [Shortcuts]]],
    indirect=True
)
@pytest.mark.skipif(
    sys.platform.startswith("linux") and running_in_ci(),
    reason="Fails on Linux and CI"
)
def test_shortcut_for_widget_is_updated(config_dialog, codeeditor, qtbot):
    """Test shortcuts for codeeditor are updated on the fly."""
    editor = codeeditor
    text = ('aa\nbb\ncc\ndd\n')
    editor.set_text(text)

    # We need to wait for a bit so shortcuts are registered correctly
    qtbot.wait(300)

    # Check shortcuts were registered
    assert editor._shortcuts != {}

    # Check "move line down" shortcut is working as expected
    qtbot.keyClick(editor, Qt.Key_Down, modifier=Qt.AltModifier)
    assert editor.toPlainText() == "bb\naa\ncc\ndd\n"

    # Change "move line down" to a different shortcut
    editor.set_conf("editor/move line down", "Ctrl+B", section="shortcuts")
    qtbot.wait(300)

    # Check new shortcut works
    qtbot.keyClick(editor, Qt.Key_B, modifier=Qt.ControlModifier)
    assert editor.toPlainText() == "bb\ncc\naa\ndd\n"

    # Check previous shortcut doesn't work
    qtbot.keyClick(editor, Qt.Key_Down, modifier=Qt.AltModifier)
    assert editor.toPlainText() == "bb\ncc\naa\ndd\n"

    # Reset all shortcuts to defaults (as users would do it)
    configpage = config_dialog.get_page()
    configpage.reset_to_default(force=True)
    qtbot.wait(300)

    # Make sure we are at the right line before the next check
    block_to_be = editor.document().findBlockByLineNumber(2)
    cursor = editor.textCursor()
    cursor.setPosition(block_to_be.position())
    editor.setTextCursor(cursor)

    # Check default shortcut works
    qtbot.keyClick(editor, Qt.Key_Down, modifier=Qt.AltModifier)
    assert editor.toPlainText() == "bb\ncc\ndd\naa\n"

    # Check new shortcut doesn't work
    qtbot.keyClick(editor, Qt.Key_B, modifier=Qt.ControlModifier)
    assert editor.toPlainText() == "bb\ncc\ndd\naa\n"


if __name__ == '__main__':
    pytest.main(['test_codeeditor.py'])



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# ----------------------------------------------------------------------------

"""CompletionPlugin tests."""

# Third party imports
import pytest
from qtpy.QtCore import QObject, Signal, Slot

# Local imports
from spyder.plugins.completion.api import (
    SpyderCompletionProvider, CompletionRequestTypes)


class DummyCompletionReceiver(QObject):
    """Dummy class that can handle LSP responses."""
    sig_response = Signal(str, dict)

    @Slot(str, dict)
    def handle_response(self, method, params):
        self.sig_response.emit(method, params)


class FakeProvider(SpyderCompletionProvider):
    COMPLETION_PROVIDER_NAME = 'fake'
    CONF_DEFAULTS = [
        ('key1', 'value1'),
        ('key2', 'value2'),
        ('key3', 'value3'),
        ('key4', 4)
    ]
    CONF_VERSION = "0.1.0"


@pytest.fixture
def completion_receiver(completion_plugin_all_started):
    completion_plugin, _ = completion_plugin_all_started
    receiver = DummyCompletionReceiver(None)
    return completion_plugin, receiver


def test_configuration_merge(completion_plugin_all):
    first_defaults = dict(FakeProvider.CONF_DEFAULTS)
    first_version = FakeProvider.CONF_VERSION

    # Check that a new completion provider configuration is registered without
    # changes
    result = completion_plugin_all._merge_default_configurations(
        FakeProvider, FakeProvider.COMPLETION_PROVIDER_NAME, {}
    )
    (conf_version, conf_values, conf_defaults) = result

    assert conf_version == first_version
    assert conf_values == first_defaults
    assert conf_defaults == first_defaults

    # Add a new value to the initial default configuration without changing the
    # version
    second_config = first_defaults.copy()
    second_config['extra_value'] = ['value']

    FakeProvider.CONF_DEFAULTS = [(k, v) for k, v in second_config.items()]

    prev_config = {
        FakeProvider.COMPLETION_PROVIDER_NAME: {
            'version': first_version,
            'values': first_defaults,
            'defaults': first_defaults
        }
    }

    result = completion_plugin_all._merge_default_configurations(
        FakeProvider, FakeProvider.COMPLETION_PROVIDER_NAME, prev_config
    )
    (conf_version, conf_values, conf_defaults) = result

    assert conf_version == first_version
    assert conf_values == second_config
    assert conf_defaults == second_config

    # Assert that default values cannot be changed without a bump in the minor
    # version
    config = first_defaults.copy()
    config['key4'] = 5

    third_config = first_defaults.copy()
    third_config['key4'] = -1

    FakeProvider.CONF_DEFAULTS = [(k, v) for k, v in third_config.items()]

    prev_config = {
        FakeProvider.COMPLETION_PROVIDER_NAME: {
            'version': first_version,
            'values': config,
            'defaults': first_defaults
        }
    }

    result = completion_plugin_all._merge_default_configurations(
        FakeProvider, FakeProvider.COMPLETION_PROVIDER_NAME, prev_config
    )
    (conf_version, conf_values, conf_defaults) = result

    assert conf_version == first_version
    assert conf_values == config
    assert conf_defaults == first_defaults

    # Assert that default values can be replaced with new ones when the
    # minor version number is bumped.
    config['key1'] = 'othervalue'
    expected_config = config.copy()
    expected_config['key4'] = -1

    FakeProvider.CONF_VERSION = "0.1.1"

    result = completion_plugin_all._merge_default_configurations(
        FakeProvider, FakeProvider.COMPLETION_PROVIDER_NAME, prev_config
    )
    (conf_version, conf_values, conf_defaults) = result

    assert conf_version == "0.1.1"
    assert conf_values == expected_config
    assert conf_defaults == third_config

    # Ensure that default values cannot be removed if the major version is not
    # bumped
    fourth_config = third_config.copy()
    fourth_config.pop('key2')

    FakeProvider.CONF_DEFAULTS = [(k, v) for k, v in fourth_config.items()]

    result = completion_plugin_all._merge_default_configurations(
        FakeProvider, FakeProvider.COMPLETION_PROVIDER_NAME, prev_config
    )
    (conf_version, conf_values, conf_defaults) = result

    assert conf_version == "0.1.1"
    assert conf_values == expected_config
    assert conf_defaults == third_config

    # Remove an option when the major version is bumped.
    FakeProvider.CONF_VERSION = "1.0.0"
    expected_config.pop('key2')

    result = completion_plugin_all._merge_default_configurations(
        FakeProvider, FakeProvider.COMPLETION_PROVIDER_NAME, prev_config
    )
    (conf_version, conf_values, conf_defaults) = result

    assert conf_version == "1.0.0"
    assert conf_values == expected_config
    assert conf_defaults == fourth_config


def test_provider_detection(completion_plugin_all):
    print(completion_plugin_all.providers)
    assert len(completion_plugin_all.providers) == 3


@pytest.mark.order(1)
def test_plugin_completion_gather(qtbot_module, completion_receiver):
    completion, receiver = completion_receiver

    # Parameters to perform a textDocument/didOpen request
    params = {
        'file': 'test.py',
        'language': 'python',
        'version': 1,
        'text': "# This is some text with some classe\nimport os\n\ncla",
        'response_instance': receiver,
        'offset': 1,
        'selection_start': 0,
        'selection_end': 0,
        'codeeditor': receiver,
        'requires_response': False
    }

    with qtbot_module.waitSignal(receiver.sig_response, timeout=30000) as blocker:
        completion.send_request(
            'python', CompletionRequestTypes.DOCUMENT_DID_OPEN, params)

    # Parameters to perform a textDocument/completion request
    params = {
        'file': 'test.py',
        'line': 2,
        'column': 3,
        'offset': 50,
        'selection_start': 0,
        'selection_end': 0,
        'current_word': 'cla',
        'codeeditor': receiver,
        'response_instance': receiver,
        'requires_response': True
    }

    with qtbot_module.waitSignal(receiver.sig_response, timeout=30000) as blocker:
        completion.send_request(
            'python', CompletionRequestTypes.DOCUMENT_COMPLETION, params)

    _, response = blocker.args

    response = response['params']
    provider_set = {x['provider'] for x in response}

    # Assert the response contains information from all the providers
    provider_set == {'LSP', 'Fallback', 'Snippets'}


@pytest.mark.order(1)
def test_plugin_first_response_request(qtbot_module, completion_receiver):
    completion, receiver = completion_receiver

    # Parameters to perform a textDocument/didOpen request
    params = {
        'file': 'test2.py',
        'language': 'python',
        'version': 2,
        'text': "# This is some text with some classe\nimport os\n\n",
        'response_instance': receiver,
        'offset': 1,
        'diff': '',
        'selection_start': 0,
        'selection_end': 0,
        'codeeditor': receiver,
        'requires_response': False
    }

    with qtbot_module.waitSignal(receiver.sig_response, timeout=30000) as blocker:
        completion.send_request(
            'python', CompletionRequestTypes.DOCUMENT_DID_OPEN, params)


    params = {
        'file': 'test2.py',
        'line': 1,
        'column': 8,
        'offset': 43,
        'diff': '',
        'response_instance': receiver,
        'codeeditor': receiver,
        'requires_response': True
    }

    with qtbot_module.waitSignal(receiver.sig_response, timeout=30000) as blocker:
        completion.send_request(
            'python', CompletionRequestTypes.DOCUMENT_HOVER, params)

    _, response = blocker.args
    assert len(response['params']) > 0



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for profiler.py
"""


# Third party imports
from qtpy.QtGui import QIcon
import pytest

# Local imports
from spyder.plugins.profiler.widgets.main_widget import ProfilerDataTree
from spyder.utils.palette import SpyderPalette


ERROR = SpyderPalette.COLOR_ERROR_1
SUCESS = SpyderPalette.COLOR_SUCCESS_1


# --- Fixtures
# -----------------------------------------------------------------------------
@pytest.fixture
def profiler_datatree_bot(qtbot):
    """Set up Profiler widget."""
    # Avoid qtawesome startup errors
    ProfilerDataTree.create_icon = lambda x, y: QIcon()
    ProfilerDataTree.CONF_SECTION = ''
    tree = ProfilerDataTree(None)
    qtbot.addWidget(tree)
    tree.show()
    yield tree
    tree.destroy()


# --- Tests
# -----------------------------------------------------------------------------
def test_format_measure(profiler_datatree_bot):
    """ Test ProfilerDataTree.format_measure()."""
    tree = profiler_datatree_bot
    fm = tree.format_measure
    assert fm(125) == '125'
    assert fm(1.25e-8) == '12.50 ns'
    assert fm(1.25e-5) == u'12.50 \u03BCs'
    assert fm(1.25e-2) == '12.50 ms'
    assert fm(12.5) == '12.50 s'
    assert fm(125.5) == '2.5 min'
    assert fm(12555.5) == '3h:29min'

    assert fm(-125) == '125'
    assert fm(-1.25e-8) == '12.50 ns'
    assert fm(-1.25e-5) == u'12.50 \u03BCs'
    assert fm(-1.25e-2) == '12.50 ms'
    assert fm(-12.5) == '12.50 s'
    assert fm(-125.5) == '2.5 min'
    assert fm(-12555.5) == '3h:29min'


def test_color_string(profiler_datatree_bot):
    """ Test ProfilerDataTree.color_string()."""
    tree = profiler_datatree_bot
    cs = tree.color_string

    tree.compare_file = 'test'
    assert cs([5.0]) == ['5.00 s', ['', 'black']]
    assert cs([1.251e-5, 1.251e-5]) == [u'12.51 \u03BCs', ['', 'black']]
    assert cs([5.0, 4.0]) == ['5.00 s', ['+1000.00 ms', ERROR]]
    assert cs([4.0, 5.0]) == ['4.00 s', ['-1000.00 ms', SUCESS]]

    tree.compare_file = None
    assert cs([4.0, 5.0]) == ['4.00 s', ['', 'black']]


def test_format_output(profiler_datatree_bot):
    """ Test ProfilerDataTree.format_output()."""
    tree = profiler_datatree_bot
    fo = tree.format_output

    # Mock Stats class to be able to use fixed data for input.
    class Stats:
        stats = {}

    tree.stats1 = [Stats(), Stats()]
    tree.stats1[0].stats = {('key1'): (1, 1000, 3.5, 1.5, {}),
                            ('key2'): (1, 1200, 2.0, 2.0, {})
                            }
    tree.stats1[1].stats = {('key1'): (1, 1000, 3.7, 1.3, {}),
                            ('key2'): (1, 1199, 2.4, 2.4, {})
                            }

    tree.compare_file = 'test'
    assert list((fo('key1'))) == [['1000', ['', 'black']],
                                  ['3.50 s', ['-200.00 ms', SUCESS]],
                                  ['1.50 s', ['+200.00 ms', ERROR]]]
    assert list((fo('key2'))) == [['1200', ['+1', ERROR]],
                                  ['2.00 s', ['-400.00 ms', SUCESS]],
                                  ['2.00 s', ['-400.00 ms', SUCESS]]]


if __name__ == "__main__":
    pytest.main()
